﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Users
{
    /// <summary>
    /// controller for the group roles
    /// </summary>
    [Route(Constants.ApiPrefix + "grouproles")]
    public class GroupRolesController : ControllerWithDb
    {
        public GroupRolesController(SyndromDbContext context) : base(context)
        {
        }

        /// <summary>
        /// returns all known group roles
        /// </summary>
        [HttpGet()]
        public async Task GetAll()
        {
            if (!await base.IsLoggedIn()) return;

            if (
                !HasSystemPermission(
                    permission => permission != null &&
                                  (
                                      permission.SystemRolePermission.CanCreateRoles
                                      || permission.SystemRolePermission.CanChangeRoles
                                      || permission.SystemRolePermission.CanDeleteRoles
                                  )))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var groupRoles = await _context.GroupRoles
                    .Include(p => p.GroupRolePermission)
                    .ToListAsync()
                ;
            
            /*
            if (
                !HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanChangeOtherUsersSystemRole))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
            }
            */


            var rolesForFrontend = new List<GroupRoleWithPermissionFullBase>();

            foreach (GroupRole groupRole in groupRoles)
            {
                rolesForFrontend.Add(new GroupRoleWithPermissionFullBase
                (
                    id: groupRole.Id,
                    displayName: groupRole.DisplayName,
                    canChangeOtherMembersRole: groupRole.GroupRolePermission.CanChangeOtherMembersRole,
                    canCreateExercises: groupRole.GroupRolePermission.CanCreateExercises,
                    canAddUserToGroup: groupRole.GroupRolePermission.CanAddUserToGroup,
                    canAssessExercises: groupRole.GroupRolePermission.CanAssessExercises,
                    canChangeExercises: groupRole.GroupRolePermission.CanChangeExercises,
                    canDeleteExercises: groupRole.GroupRolePermission.CanDeleteExercises,
                    canLockExercisesPermanently: groupRole.GroupRolePermission.CanLockExercisesPermanently,
                    canReleaseExercises: groupRole.GroupRolePermission.CanReleaseExercises,
                    canRemoveMemberFromGroup: groupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                    canSeeExercisesFromOthersInGroup: groupRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup,
                    canSeeOtherMembers: groupRole.GroupRolePermission.CanSeeOtherMembers,
                    canChangeGroupData: groupRole.GroupRolePermission.CanChangeGroupData
                ));
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<GroupRoleWithPermissionFullBase>>(ResponseCode.Ok,
                        "", rolesForFrontend)));
        }

        /// <summary>
        /// returns all known group roles but checks the group permission not the system permission
        /// </summary>
        [HttpGet("{groupId}/roles")]
        public async Task GetAllGroupRoles(int groupId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //user needs to be member of the group to see the members...
            UserWithUserGroup connection = await _context.UserWithUserGroups
                    .FirstOrDefaultAsync(p => p.UserId == userId && p.UserGroupId == groupId)
                ;

            if (connection == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.Ok, "user group (for roles) not found")));
                return;
            }

            if (
                !await HasGroupPermission(groupId,
                    permission => permission != null &&
                                  (
                                      permission.CanSeeOtherMembers
                                  )))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var groupRoles = await _context.GroupRoles
                    .Include(p => p.GroupRolePermission)
                    .ToListAsync()
                ;

            var rolesForFrontend = new List<GroupRoleWithPermissionFullBase>();

            foreach (GroupRole groupRole in groupRoles)
            {
                rolesForFrontend.Add(new GroupRoleWithPermissionFullBase
                (
                    id: groupRole.Id,
                    displayName: groupRole.DisplayName,
                    canChangeOtherMembersRole: groupRole.GroupRolePermission.CanChangeOtherMembersRole,
                    canCreateExercises: groupRole.GroupRolePermission.CanCreateExercises,
                    canAddUserToGroup: groupRole.GroupRolePermission.CanAddUserToGroup,
                    canAssessExercises: groupRole.GroupRolePermission.CanAssessExercises,
                    canChangeExercises: groupRole.GroupRolePermission.CanChangeExercises,
                    canDeleteExercises: groupRole.GroupRolePermission.CanDeleteExercises,
                    canLockExercisesPermanently: groupRole.GroupRolePermission.CanLockExercisesPermanently,
                    canReleaseExercises: groupRole.GroupRolePermission.CanReleaseExercises,
                    canRemoveMemberFromGroup: groupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                    canSeeExercisesFromOthersInGroup: groupRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup,
                    canSeeOtherMembers: groupRole.GroupRolePermission.CanSeeOtherMembers,
                    canChangeGroupData: groupRole.GroupRolePermission.CanChangeGroupData
                ));
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<GroupRoleWithPermissionFullBase>>(ResponseCode.Ok,
                        "group roles queried", rolesForFrontend)));
        }


        /// <summary>
        /// creates a new group role
        /// </summary>
        /// <param name="role">the group role to create</param>
        [HttpPost("create")]
        public async Task Create([FromBody] GroupRoleWithPermissionFullBase role)
        {
            if (!await base.IsLoggedIn()) return;


            if (
                !HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanCreateRoles))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var newRole = new GroupRole()
            {
                DisplayName = role.DisplayName.Trim(),
                GroupRolePermission = new GroupRolePermission
                (
                    CanAddUserToGroup: role.CanAddUserToGroup,
                    CanAssessExercises: role.CanAssessExercises,
                    CanChangeOtherMembersRole: role.CanChangeOtherMembersRole,
                    CanCreateExercises: role.CanCreateExercises,
                    CanDeleteExercises: role.CanDeleteExercises,
                    CanLockExercisesPermanently: role.CanLockExercisesPermanently,
                    CanSeeExercisesFromOthersInGroup: role.CanSeeExercisesFromOthersInGroup,
                    CanRemoveMemberFromGroup: role.CanRemoveMemberFromGroup,
                    CanChangeExercises: role.CanChangeExercises,
                    CanReleaseExercises: role.CanReleaseExercises,
                    CanSeeOtherMembers: role.CanSeeExercisesFromOthersInGroup,
                    CanChangeGroupData: role.CanChangeGroupData
                )
            };

            try
            {
                _context.GroupRoles.Add(newRole);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }


            var groupRoleForFrontend = new GroupRoleWithPermissionFullBase
            (
                id: newRole.Id,
                displayName: newRole.DisplayName,
                canChangeOtherMembersRole: newRole.GroupRolePermission.CanChangeOtherMembersRole,
                canChangeExercises: newRole.GroupRolePermission.CanChangeExercises,
                canCreateExercises: newRole.GroupRolePermission.CanCreateExercises,
                canRemoveMemberFromGroup: newRole.GroupRolePermission.CanRemoveMemberFromGroup,
                canReleaseExercises: newRole.GroupRolePermission.CanReleaseExercises,
                canAddUserToGroup: newRole.GroupRolePermission.CanAddUserToGroup,
                canDeleteExercises: newRole.GroupRolePermission.CanDeleteExercises,
                canLockExercisesPermanently: newRole.GroupRolePermission.CanLockExercisesPermanently,
                canAssessExercises: newRole.GroupRolePermission.CanAssessExercises,
                canSeeExercisesFromOthersInGroup: newRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup,
                canSeeOtherMembers: newRole.GroupRolePermission.CanSeeOtherMembers,
                canChangeGroupData: newRole.GroupRolePermission.CanChangeGroupData
            );


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<GroupRoleWithPermissionFullBase>(ResponseCode.Ok,
                        "created group role", groupRoleForFrontend)));
        }

        /// <summary>
        /// changes a group role
        /// </summary>
        /// <param name="role">the group role to change</param>
        [HttpPost("update")]
        public async Task Update([FromBody] GroupRoleWithPermissionFullBase role)
        {
            if (!await base.IsLoggedIn()) return;


            if (
                !HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanChangeRoles))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            GroupRole oldRole = await _context.GroupRoles
                .Include(p => p.GroupRolePermission)
                .FirstOrDefaultAsync(p => p.Id == role.Id);


            if (oldRole == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group role not found")));
                return;
            }

            oldRole.DisplayName = role.DisplayName.Trim();
            oldRole.GroupRolePermission.CanAddUserToGroup = role.CanAddUserToGroup;
            oldRole.GroupRolePermission.CanAssessExercises = role.CanAssessExercises;
            oldRole.GroupRolePermission.CanChangeOtherMembersRole = role.CanChangeOtherMembersRole;
            oldRole.GroupRolePermission.CanCreateExercises = role.CanCreateExercises;
            oldRole.GroupRolePermission.CanDeleteExercises = role.CanDeleteExercises;
            oldRole.GroupRolePermission.CanLockExercisesPermanently = role.CanLockExercisesPermanently;
            oldRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup = role.CanSeeExercisesFromOthersInGroup;
            oldRole.GroupRolePermission.CanRemoveMemberFromGroup = role.CanRemoveMemberFromGroup;
            oldRole.GroupRolePermission.CanChangeExercises = role.CanChangeExercises;
            oldRole.GroupRolePermission.CanReleaseExercises = role.CanReleaseExercises;
            oldRole.GroupRolePermission.CanSeeOtherMembers = role.CanSeeOtherMembers;
            oldRole.GroupRolePermission.CanChangeGroupData = role.CanChangeGroupData;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            var groupRoleForFrontend = new GroupRoleWithPermissionFullBase
            (
                id: oldRole.Id,
                displayName: oldRole.DisplayName,
                canChangeOtherMembersRole: oldRole.GroupRolePermission.CanChangeOtherMembersRole,
                canChangeExercises: oldRole.GroupRolePermission.CanChangeExercises,
                canCreateExercises: oldRole.GroupRolePermission.CanCreateExercises,
                canRemoveMemberFromGroup: oldRole.GroupRolePermission.CanRemoveMemberFromGroup,
                canReleaseExercises: oldRole.GroupRolePermission.CanReleaseExercises,
                canAddUserToGroup: oldRole.GroupRolePermission.CanAddUserToGroup,
                canDeleteExercises: oldRole.GroupRolePermission.CanDeleteExercises,
                canLockExercisesPermanently: oldRole.GroupRolePermission.CanLockExercisesPermanently,
                canAssessExercises: oldRole.GroupRolePermission.CanAssessExercises,
                canSeeExercisesFromOthersInGroup: oldRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup,
                canSeeOtherMembers: oldRole.GroupRolePermission.CanSeeOtherMembers,
                canChangeGroupData: oldRole.GroupRolePermission.CanChangeGroupData
            );

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<GroupRoleWithPermissionFullBase>(ResponseCode.Ok,
                        "group role changed", groupRoleForFrontend)));
        }


        /// <summary>
        /// removes the given group role
        /// </summary>
        /// <param name="id">the group role</param>
        [HttpDelete("delete/{id}")]
        public async Task Delete(int id)
        {
            if (!await base.IsLoggedIn()) return;

            if (
                !HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanDeleteRoles))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var systemSettings = await _context.SystemSettings
                .FirstOrDefaultAsync();

            if (systemSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "no system settings found")));
                return;
            }

            if (systemSettings.DefaultGroupRoleId == id)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "cannot delete the system default group role")));
                return;
            }
            
            if (systemSettings.DefaultGroupCreatorGroupRoleId == id)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "cannot delete the system default group creator role")));
                return;
            }


            GroupRole oldRole = await _context.GroupRoles
                .FirstOrDefaultAsync(p => p.Id == id);


            if (oldRole == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group role not found")));
                return;
            }

            //very user needs a group role...
            //when deleting the role change all users role to system default one because they would have no group role...

            List<UserWithUserGroup> conenctions =
                await _context.UserWithUserGroups
                    .Where(p => p.GroupRoleId == oldRole.Id)
                    .ToListAsync();

            foreach (var userWithUserGroup in conenctions)
            {
                userWithUserGroup.GroupRoleId = systemSettings.DefaultGroupRoleId;
            }


            try
            {
                _context.GroupRoles.Remove(oldRole);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
               await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "grou role deleted")));
        }
    }


    public class GroupRoleFullBase
    {
        public int Id { get; set; }
        
        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }
    }
    
    /// <summary>
    /// the group roles woth permissions
    /// </summary>
    public class GroupRoleWithPermissionFullBase : GroupRoleFullBase
    {

        public GroupRoleWithPermissionFullBase(int id, string displayName, bool canCreateExercises,
            bool canChangeExercises, bool canDeleteExercises, bool canLockExercisesPermanently, bool canSeeExercisesFromOthersInGroup,
            bool canReleaseExercises, bool canAssessExercises, bool canAddUserToGroup,
            bool canRemoveMemberFromGroup, bool canChangeOtherMembersRole, bool canSeeOtherMembers,
            bool canChangeGroupData)
        {
            this.Id = id;
            this.DisplayName = displayName;
            this.CanCreateExercises = canCreateExercises;
            this.CanChangeExercises = canChangeExercises;
            this.CanDeleteExercises = canDeleteExercises;
            this.CanLockExercisesPermanently = canLockExercisesPermanently;
            this.CanSeeExercisesFromOthersInGroup = canSeeExercisesFromOthersInGroup;
            this.CanReleaseExercises = canReleaseExercises;
            this.CanAssessExercises = canAssessExercises;
            this.CanAddUserToGroup = canAddUserToGroup;
            this.CanRemoveMemberFromGroup = canRemoveMemberFromGroup;
            this.CanChangeOtherMembersRole = canChangeOtherMembersRole;
            this.CanSeeOtherMembers = canSeeOtherMembers;
            this.CanChangeGroupData = canChangeGroupData;
        }



        /// <summary>
        /// trze: can create new exercises, fale: not
        /// </summary>
        public bool CanCreateExercises { get; set; }

        /// <summary>
        /// true: can change exercises, false: not
        /// </summary>
        public bool CanChangeExercises { get; set; }

        /// <summary>
        /// true: can delete exercises, false: not
        /// </summary>
        public bool CanDeleteExercises { get; set; }


        /// <summary>
        /// true: can see exercises from other users in the group, false: not
        /// </summary>
        public bool CanSeeExercisesFromOthersInGroup { get; set; }

        /// <summary>
        /// true: can release exercises, false: not
        /// </summary>
        public bool CanReleaseExercises { get; set; }

        /// <summary>
        /// true: can copy exercises, false: not
        /// </summary>
        public bool CanCopyExercises { get; set; }
        
        /// <summary>
        /// true: can lock exercises, false not
        /// </summary>
        public bool CanLockExercisesPermanently { get; set; }

        /// <summary>
        /// true: can assess/evaulate exercises manually, false: not
        /// </summary>
        public bool CanAssessExercises { get; set; }

        /// <summary>
        /// true: can add a person to the group, false: not
        /// </summary>
        public bool CanAddUserToGroup { get; set; }

        /// <summary>
        /// true: can remove a person from a group, false: not
        /// </summary>
        public bool CanRemoveMemberFromGroup { get; set; }

        /// <summary>
        /// true: can change a role of other users in the group, false: not
        /// </summary>
        public bool CanChangeOtherMembersRole { get; set; }

        /// <summary>
        /// true: can see other members in single group overview (and get all group roles)
        /// false: not
        /// </summary>
        public bool CanSeeOtherMembers { get; set; }

        /// <summary>
        /// true: can change the group data e.g. name, email
        /// false: not
        /// </summary>
        public bool CanChangeGroupData { get; set; }
    }
}
